﻿using PdfSharp.Drawing;
using PdfSharp.Pdf;
using System;
using System.Collections.Generic;
using System.Data;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;

namespace PdfHelperFW
{
    public class PdfHelper
    {
        readonly DataTable _table;
        readonly int _timestampColIdx = 0;

        // Page structure
        readonly int PAGE_HEIGHT = 842; // A4
        readonly int PAGE_WIDTH = 595; // A4

        readonly string FONT_NAME = "Malgun Gothic";
        readonly double FONT_SIZE = 5.5d;

        readonly int MARGIN_LEFT = 20;
        readonly int MARGIN_TOP = 45;

        readonly bool SORT_ASC = true;

        // Elements
        readonly bool _enableNoCol = true;
        readonly double NO_COL_MARGIN = 4;

        readonly double CONTENT_WIDTH = 555;
        readonly double TIMESTAMP_CELL_WIDTH = 70;

        public readonly double MAX_CELL_COLUMN_COUNT = 10;
        readonly double ALL_CELL_WIDTH = 485;
        readonly double CELL_WIDTH = 48.5;
        readonly int CELL_HEIGHT = 15;

        // Design
        readonly XSolidBrush HEADER_RECT_STYLE = new XSolidBrush(XColor.FromArgb(210, 51, 52));
        readonly XPen RECT_EDGE_PEN = new XPen(XColor.FromArgb(34, 30, 32), 0.25d);
        readonly XBrush HEADER_STRING_COLOR = new XSolidBrush(XColors.White);
        readonly XBrush STRING_COLOR = new XSolidBrush(XColor.FromArgb(34, 30, 32));

        readonly int ROW_COUNT_PER_PAGE = 50;

        PdfDocument _document;
        XFont _font;
        XFont _boldFont;
        XStringFormat _format;

        // Logo
        readonly XImage LOGO_IMAGE;
        readonly double LOGO_WIDTH;
        readonly double LOGO_HEIGHT;
        readonly double LOGO_X;
        readonly double LOGO_Y;
        readonly double LOGO_Y_MARGIN = 5;
        MemoryStream _ms;

        public PdfHelper(DataTable table, int timestampColIdx = 0, bool enableNoCol = true, bool sortAsc = true)
        {
            if (table.Columns.Count - 1 > MAX_CELL_COLUMN_COUNT)
                throw new ArgumentException($"The number of data should be less than {MAX_CELL_COLUMN_COUNT}");

            _table = table;
            _timestampColIdx = timestampColIdx;

            _enableNoCol = enableNoCol;

            SORT_ASC = sortAsc;

            // Check cell width
            CELL_WIDTH = ALL_CELL_WIDTH / (table.Columns.Count - 1);

            _document = new PdfDocument();
            _document.Info.Title = "NS Report";
            _document.Info.Author = "Particle Measuring Systems Korea";

            _font = new XFont(FONT_NAME, FONT_SIZE, XFontStyle.Regular);
            _boldFont = new XFont(FONT_NAME, FONT_SIZE, XFontStyle.Bold);
            _format = new XStringFormat();
            _format.LineAlignment = XLineAlignment.Center;
            _format.Alignment = XStringAlignment.Center;
        }

        public PdfHelper(DataTable table, string logoPath, double logoHeight, int timestampColIdx = 0, bool enableNoCol = true, bool sortAsc = true) : this(table, timestampColIdx, enableNoCol, sortAsc)
        {
            if (string.IsNullOrEmpty(logoPath) || !File.Exists(logoPath) ||
                logoHeight >= MARGIN_TOP - LOGO_Y_MARGIN)
                throw new ArgumentException($"Some arguments for logo are invalid");

            LOGO_IMAGE = XImage.FromFile(logoPath);
            LOGO_HEIGHT = logoHeight;
            LOGO_WIDTH = LOGO_IMAGE.PointWidth * (LOGO_HEIGHT / LOGO_IMAGE.PointHeight);
            LOGO_X = MARGIN_LEFT + CONTENT_WIDTH - LOGO_WIDTH;
            LOGO_Y = MARGIN_TOP - LOGO_HEIGHT - LOGO_Y_MARGIN;
        }

        public PdfHelper(DataTable table, Bitmap logoBitmap, double logoHeight, int timestampColIdx = 0, bool enableNoCol = true, bool sortAsc = true) : this(table, timestampColIdx, enableNoCol, sortAsc)
        {
            if (logoBitmap == null
                || logoHeight >= MARGIN_TOP - LOGO_Y_MARGIN)
                throw new ArgumentException($"Some arguments for logo are invalid");

            _ms = new MemoryStream();

            logoBitmap.Save(_ms, System.Drawing.Imaging.ImageFormat.Jpeg);
            _ms.Position = 0;

            LOGO_IMAGE = XImage.FromStream(_ms);

            LOGO_HEIGHT = logoHeight;
            LOGO_WIDTH = LOGO_IMAGE.PointWidth * (LOGO_HEIGHT / LOGO_IMAGE.PointHeight);
            LOGO_X = MARGIN_LEFT + CONTENT_WIDTH - LOGO_WIDTH;
            LOGO_Y = MARGIN_TOP - LOGO_HEIGHT - LOGO_Y_MARGIN;
        }

        public void Save(string targetPath)
        {
            EnumerableRowCollection sorted;
            if (SORT_ASC)
                sorted = _table.AsEnumerable().OrderBy(row => row.Field<DateTime>(_timestampColIdx));
            else
                sorted = _table.AsEnumerable().OrderByDescending(row => row.Field<DateTime>(_timestampColIdx));

            DataTable sortedTable = _table.Clone();
            foreach (DataRow row in sorted)
            {
                sortedTable.ImportRow(row);
            }

            for (int startRow = 0; startRow < sortedTable.Rows.Count; startRow += ROW_COUNT_PER_PAGE)
            {
                PdfPage page = _document.AddPage();
                page.Size = PdfSharp.PageSize.A4;
                XGraphics xg = XGraphics.FromPdfPage(page);

                // Logo
                if (LOGO_IMAGE != null)
                    xg.DrawImage(LOGO_IMAGE, LOGO_X, LOGO_Y, LOGO_WIDTH, LOGO_HEIGHT);

                // Header
                double x = MARGIN_LEFT;
                double y = MARGIN_TOP;
                int endRow = Math.Min(startRow + ROW_COUNT_PER_PAGE, sortedTable.Rows.Count);
                double cellWidth = CELL_WIDTH;
                double noCellWidth = 0;
                // Cell width recalculation
                if (_enableNoCol)
                {
                    int maxNo = endRow + 1;
                    noCellWidth = MeasureStringWidth(xg, maxNo.ToString(), _font) + NO_COL_MARGIN;
                    cellWidth = (ALL_CELL_WIDTH - noCellWidth) / (sortedTable.Columns.Count - 1);

                    XRect rect = new XRect(x, y, noCellWidth, CELL_HEIGHT);
                    xg.DrawRectangle(HEADER_RECT_STYLE, rect);
                    xg.DrawRectangle(RECT_EDGE_PEN, rect);
                    //xg.DrawString("No.", _boldFont, HEADER_STRING_COLOR, rect, _format);

                    x += noCellWidth;
                }

                for (int col = 0; col < sortedTable.Columns.Count; col++)
                {
                    double increment = 0;
                    XRect rect;
                    if (col == _timestampColIdx)
                    {
                        rect = new XRect(x, y, TIMESTAMP_CELL_WIDTH, CELL_HEIGHT);
                        increment = TIMESTAMP_CELL_WIDTH;
                    }
                    else
                    {
                        rect = new XRect(x, y, cellWidth, CELL_HEIGHT);
                        increment = cellWidth;
                    }

                    xg.DrawRectangle(HEADER_RECT_STYLE, rect);
                    xg.DrawRectangle(RECT_EDGE_PEN, rect);
                    xg.DrawString(sortedTable.Columns[col].ColumnName, _boldFont, HEADER_STRING_COLOR, rect, _format);

                    x += increment;
                }

                // Row
                for (int row = startRow; row < endRow; row++)
                {
                    x = MARGIN_LEFT;
                    y += CELL_HEIGHT;

                    if (_enableNoCol)
                    {
                        XRect rect = new XRect(x, y, noCellWidth, CELL_HEIGHT);
                        xg.DrawRectangle(RECT_EDGE_PEN, rect);
                        xg.DrawString((row + 1).ToString(), _font, STRING_COLOR, rect, _format);
                        x += noCellWidth;
                    }
                    
                    for (int col = 0; col < sortedTable.Columns.Count; col++)
                    {
                        double increment = 0;
                        XRect rect;
                        string value;
                        if (col == _timestampColIdx)
                        {
                            rect = new XRect(x, y, TIMESTAMP_CELL_WIDTH, CELL_HEIGHT);
                            DateTime date = (DateTime)sortedTable.Rows[row][col];
                            value = date.ToString("yyyy-MM-dd HH:mm:ss");
                            increment = TIMESTAMP_CELL_WIDTH;
                        }
                        else
                        {
                            rect = new XRect(x, y, cellWidth, CELL_HEIGHT);
                            value = sortedTable.Rows[row][col].ToString();
                            increment = cellWidth;
                        }

                        xg.DrawRectangle(RECT_EDGE_PEN, rect);
                        xg.DrawString(value, _font, STRING_COLOR, rect, _format);

                        x += increment;
                    }
                }
            }

            _document.Save(targetPath);

            if (_ms != null)
                _ms.Dispose();
        }

        double MeasureStringWidth(XGraphics xg, string text, XFont font)
        {
            XSize size = xg.MeasureString(text, font);
            return size.Width;
        }
    }
}
